// Author: Krzysztof Onak
// Part of Homework 11 for DS-210 at Boston University (Spring 2022)
//
// It contains a correct and incorrect implementation of Dijkstra's
// algorithm.
//
use std::cmp::Reverse;
use std::collections::BinaryHeap;

// some size limits to make sure a counterexample does not overflow
// integers
const MAX_N: usize = 1000;
const MAX_EDGE_LENGTH: usize = 1000;

// useful type aliases
type Vertex = usize;
type Distance = usize;
type Edge = (Vertex, Vertex, Distance);

// a vector of distances
type DistanceVector = Vec<Option<Distance>>;

// an edge going out of a vertex, field `vertex`
// is the destination
struct Outedge {
    vertex: Vertex,
    length: Distance,
}

// list of edges going out of a vertex
type AdjacencyList = Vec<Outedge>;

// a representation of a graph with vertices labelled 0..n-1
// * field n: the number of vertices
// * field outedges: a vector of adjacency lists
struct Graph {
    n: usize,
    outedges: Vec<AdjacencyList>,
}

impl Graph {
    // create a directed graph on n vertices from a list of edges 
    fn create_directed(n: usize, edges: &Vec<Edge>) -> Graph {
        let mut outedges = Vec::<Vec<Outedge>>::with_capacity(n);
        for _ in 0..n {
            outedges.push(Vec::new());
        }
        for (u, v, length) in edges {
            outedges[*u].push(Outedge {
                vertex: *v,
                length: *length,
            });
        }
        Graph { n, outedges }
    }
}

// verification that the input data follows the required constraints
fn verify_data(n: usize, edges: &Vec<Edge>) {
    if n > MAX_N {
        panic!("Incorrect data: Too many vertices!");
    }
    for (u, v, length) in edges {
        if *u >= n || *v >= n {
            panic!("Incorrect data: Vertex label out of range");
        }
        if *length >= MAX_EDGE_LENGTH {
            panic!("Incorrect data: Edge too long!");
        }
    }
}

// An incorrect implementation of Dijkstra's algorithm: note that we remove
// the vertex at the greatest distance known distance from the already
// processed vertices.
fn incorrect_implementation(graph: &Graph, start: Vertex) -> DistanceVector {
    let mut distances = vec![None; graph.n];
    distances[start] = Some(0);

    let mut pq = BinaryHeap::new();
    pq.push((0, start));

    while let Some((dist, v)) = pq.pop() {
        if distances[v].unwrap() == dist {
            for Outedge { vertex, length } in graph.outedges[v].iter() {
                let new_dist = dist + *length;
                let update = match distances[*vertex] {
                    None => true,
                    Some(d) => new_dist < d,
                };
                if update {
                    distances[*vertex] = Some(new_dist);
                    pq.push((new_dist, *vertex));
                }
            }
        }
    }
    distances
}

// A correct implementation of Dijkstra's algorithm: here we reverse the ordering
// of tuples so that vertices at the closest distance to those whose distance
// has already been established are processed first.
fn correct_implementation(graph: &Graph, start: Vertex) -> DistanceVector {
    let mut distances = vec![None; graph.n];
    distances[start] = Some(0);

    let mut pq = BinaryHeap::new();
    pq.push(Reverse((0, start)));

    while let Some(Reverse((dist, v))) = pq.pop() {
        if distances[v].unwrap() == dist {
            for Outedge { vertex, length } in graph.outedges[v].iter() {
                let new_dist = dist + *length;
                let update = match distances[*vertex] {
                    None => true,
                    Some(d) => new_dist < d,
                };
                if update {
                    distances[*vertex] = Some(new_dist);
                    pq.push(Reverse((new_dist, *vertex)));
                }
            }
        }
    }
    distances
}

// module that should contain a counterexample
mod sample_graph;

fn main() {
    let n = sample_graph::number_of_vertices();
    let edges = sample_graph::edges();

    verify_data(n, &edges);
    let graph = Graph::create_directed(n, &edges);

    let incorrect = incorrect_implementation(&graph, 0);
    let correct = correct_implementation(&graph, 0);

    let mut same = true;
    for i in 0..n {
        if incorrect[i] != correct[i] {
            same = false;
            println!("Different distances have been computed for vertex {}", i);
            break;
        }
    }
    if same {
        println!("Both algorithms computed the same set of distances");
    }
}
